#!/usr/bin/env python3
# -*- coding:utf-8 -*-

# -----------
# SPDX-License-Identifier: MIT
# Copyright (c) 2021 Troy Williams

# uuid       = c3207508-b7ef-11eb-89b8-59b62252b88f
# author     = Troy Williams
# email      = troy.williams@bluebill.net
# date       = 2021-05-18
# -----------

"""
"""

# ------------
# System Modules - Included with Python

import sys
import logging

# ------------
# 3rd Party - From pip

import click

# ------------
# Custom Modules

from .common import run_cmd, write_json

# -------------
# Logging

# Assign the variable
log = logging.getLogger(__name__)
# -------------


@click.command("save")
@click.option(
    "--log", is_flag=True, help="Display the list of saved windows to STDOUT."
)
@click.pass_context
def save(*args, **kwargs):
    """

    # Usage

    """

    # Extract the configuration file from the click context
    paths = args[0].obj["paths"]

    # We will use the wmctrl program to save the active window positions
    # $ sudo apt install wmctrl

    # to capture the information we will use `$ wmctrl -Gl`

    # $ man wmctrl, -l:
    # List the windows being managed by the window manager. One line is output for each window, with the line broken
    # up into space separated columns.
    # - The first column always contains the window identity as a hexadecimal integer
    # - the second column always contains the desktop number (a -1 is used to identify a  sticky window)
    # - If the -p option is specified the next column will contain the PID for the window as a decimal integer
    # - If the -G option is specified then  four  integer  columns  will  follow:
    #   - x-offset
    #   - y-offset
    #   - width
    #   - height
    # - The next column always contains the client machine name. The remainder of the line contains the window title (possibly with multiple spaces in the title).

    cmd = ["wmctrl", "-lG"]

    results = run_cmd(cmd)

    # positions = [p.split()[:6] for p in results]
    items = [p.split() for p in results]

    if kwargs["log"]:

        for p in items:
            log.info(p)

    write_json(paths["locations"], items)

    if kwargs["log"]:

        log.info(f'Positions saved to {paths["locations"]}...')

    # NOTE: wmctrl -lG returns x and y in desktop coordinates, we have to multiply by the x and y scale factors to get the correct coordinates.



#!/usr/bin/env python3
import subprocess
import sys

# https://askubuntu.com/questions/702071/move-windows-to-specific-screens-using-the-command-line

# just a helper function, to reduce the amount of code
get = lambda cmd: subprocess.check_output(cmd).decode("utf-8")

# get the data on all currently connected screens, their x-resolution
screendata = [l.split() for l in get(["xrandr"]).splitlines() if " connected" in l]
screendata = sum(
    [[(w[0], s.split("+")[-2]) for s in w if s.count("+") == 2] for w in screendata], []
)

print(screendata)

# def get_class(classname):
#     # function to get all windows that belong to a specific window class (application)
#     w_list = [l.split()[0] for l in get(["wmctrl", "-l"]).splitlines()]
#     return [w for w in w_list if classname in get(["xprop", "-id", w])]

# scr = sys.argv[2]

# try:
#     # determine the left position of the targeted screen (x)
#     pos = [sc for sc in screendata if sc[0] == scr][0]
# except IndexError:
#     # warning if the screen's name is incorrect (does not exist)
#     print(scr, "does not exist. Check the screen name")
# else:
#     for w in get_class(sys.argv[1]):
#         # first move and resize the window, to make sure it fits completely inside the targeted screen
#         # else the next command will fail...
#         subprocess.Popen(["wmctrl", "-ir", w, "-e", "0,"+str(int(pos[1])+100)+",100,300,300"])
#         # maximize the window on its new screen
#         subprocess.Popen(["xdotool", "windowsize", "-sync", w, "100%", "100%"])



#!/usr/bin/env python3
# -*- coding:utf-8 -*-

# -----------
# SPDX-License-Identifier: MIT
# Copyright (c) 2021 Troy Williams

# uuid       = f6792b10-b7f5-11eb-89b8-59b62252b88f
# author     = Troy Williams
# email      = troy.williams@bluebill.net
# date       = 2021-05-18
# -----------

"""
"""

# ------------
# System Modules - Included with Python

import sys
import logging
import math

# ------------
# 3rd Party - From pip

import click

# ------------
# Custom Modules

from .common import run_cmd, read_json

# ------------
# Custom Modules


# -------------
# Logging

log = logging.getLogger(__name__)
# -------------


def move_window(wid, x, y, w, h, title, verbose=False):
    """
    Move the window specified by is (wid) to the new x, y offset and
    the new width (w) and height (h)

    $ wmctrl -ir {wid} -e 0,{x},{y},{w},{h}

    # Parameters

    wid:str
        - The id of the window to move

    x:int
        - x-coordinate of the upper left corner

    y:int
        - y-coordinate of the upper left corner

    w:int
        - width of the window

    h:int
        - height of the window

    title:str
        - The title information of the window

    verbose:bool
        - used to indicate we want more verbose information

    """

    if verbose:
        log.info(f'wmctrl -ir {wid} -e 0,{x},{y},{w},{h} "  -> {title}"')

    # restore the correct size and position on the desktop
    results = run_cmd(
        [
            "wmctrl",
            "-ir",
            wid,
            "-e",
            f"0,{x},{y},{w},{h}",
        ]
    )

    for r in results:
        log.info(r)


def set_desktop(wid, deskid, title, verbose=False):
    """
    Given the window (wid) move it to the virtual desktop (deskid)

    $ wmctrl -ir {wid} -t {deskid}

    # Parameters

    wid:str
        - The id of the window to move

    title:str
        - The title information of the window

    verbose:bool
        - used to indicate we want more verbose information
    """

    if verbose:
        log.info(f'wmctrl -ir {wid} -t {deskid} " -> {title}"')

    results = run_cmd(
        [
            "wmctrl",
            "-ir",
            wid,
            "-t",
            deskid,
        ]
    )

    for r in results:
        log.info(r)


def position_adjustments(x, y, scale_x, scale_y, title):
    """

    Given the window x,y corner offset and the window title,
    determine if the coordinates need to be adjusted.

    # Parameters

    x:int
        - x-coordinate of the upper left corner

    y:int
        - y-coordinate of the upper left corner

    scale_x:int
        - The scale factor to apply to the x coordinates in some cases

    scale_y:int
        - The scale factor to apply to the y coordinates in some cases

    title:str
        - The title information of the window

    """

    if "Firefox" in title:

        # apply a kludge to get it to restore properly...

        x -= 7
        y -= 8

    elif "Sublime Text" in title:

        # apply a kludge to get it to restore properly...

        x -= 10
        y -= 82

    elif "Discord" in title:

        # apply a kludge to get it to restore properly...

        x -= 10
        # y = math.floor(y*scale_y) + 121

    else:

        x = math.floor(x * scale_x)
        y = math.floor(y * scale_y)

    return x, y


@click.command("restore")
@click.option("--verbose", is_flag=True, help="Display more verbose output.")
@click.pass_context
def restore(*args, **kwargs):
    """

    # Usage

    """

    # Extract the configuration file from the click context
    paths = args[0].obj["paths"]

    items = read_json(paths["locations"])

    if items is None:
        log.info("Nothing to restore!")

        return

    # ----------------------
    # JSON data structure
    #   [
    #   "0x02600008",  # Window ID (Hex)
    #   "2",           # Desktop number
    #   "70",          # x-offset
    #   "460",         # y-offset
    #   "940",         # width
    #   "500",         # height
    #   "vidar",
    #   "\ud83c\udfe1-parents-only",
    #   "-",
    #   "Discord"
    # ]

    # Determine the scale factors for the x and y position
    # The scale factor will help restore the windows to the correct position given the system has
    # 2 monitors at different resolutions and the system thinks we have 5760x2160 virtual desktop

    # monitor 1: 1920 x 1080
    # monitor 2: 3840 x 2160

    m1 = [1920, 1080]
    m2 = [3840, 2160]

    scale_x = m1[0] / m2[0]
    scale_y = m1[1] / m2[1]

    for p in items:
        wid, deskid, x, y, w, h, *title = p

        title = " ".join(title)

        x, y = position_adjustments(int(x), int(y), scale_x, scale_y, title)

        # --------------
        # 1. move to the correct position on the active desktop

        move_window(wid, x, y, w, h, title, verbose=kwargs["verbose"])

        # --------------
        # 2. move to the correct desktop once it is in the correct position

        set_desktop(wid, deskid, title, verbose=kwargs["verbose"])

    # We are going to use `$ wmctrl -ir wid -e 0,x,y,w,h` to restore the window position

    # ------------
    # -i
    # Interpret  window  arguments  (<WIN>) as a numeric value rather than a string name for the window. If the nu‐
    # meric value starts with the prefix '0x' it is assumed to be a hexadecimal number.

    # -r <WIN>
    #     Specify a target window for an action.

    # -e <MVARG>
    # Resize and move a window that has been specified with a -r action according to the <MVARG> argument.

    # <MVARG>
    # A move and resize argument has the format 'g,x,y,w,h'.  All five components are integers. The first value, g,
    # is the gravity of the window, with 0 being the most common value (the default value for the  window).  Please
    # see the EWMH specification for other values.

    # The  four  remaining values are a standard geometry specification: x,y is the position of the top left corner
    # of the window, and w,h is the width and height of the window, with the exception that the value of -1 in  any
    # position is interpreted to mean that the current geometry value should not be modified.
    # ------------



#!/usr/bin/env python3
# -*- coding:utf-8 -*-

# -----------
# SPDX-License-Identifier: MIT
# Copyright (c) <year> <copyright holders>

# uuid       =
# author     = Troy Williams
# email      = troy.williams@bluebill.net
# date       =
# -----------

"""
This is a template python script for use in creating other classes and applying
a standard to them
"""

# ------------
# System Modules - Included with Python

import sys
import logging
import subprocess
import json

# ------------
# 3rd Party - From pip

# ------------
# Custom Modules


# -------------
# Logging

# Assign the variable
log = logging.getLogger(__name__)
# -------------


def write_json(path, items):
    """

    Write the items to a .json file.

    # Parameters

    path - pathlib.Path
        - the path to the file to write the data too

    items - object
        - data structure to serialize to json

    # NOTE

    It is expected that items contains objects that json can serialize

    """

    with path.open("w", encoding="utf-8") as fo:
        print(json.dumps(items, sort_keys=False, indent=2), file=fo)


def read_json(path):
    """
    Read items from a .json file, returning python objects. This is
    the pair of the "write_json" method.

    # Parameters

    path - pathlib.Path
        - the path to the file to write the date to

    # Note

    The deserialized objects may need further processing

    """

    # the path isn't set or doesn't exist...
    if path is None or not path.exists():
        return None

    with path.open("r", encoding="utf-8") as f:
        return json.loads(f.read())


def run_cmd(cmd, **kwargs):
    """
    Takes the list of arguments, cmd, and executes them via subprocess. It prints
    stdout to the terminal to report on progress or issues.

    # Parameters

    cmd: list[str]
        - The command and list of switches to execute

    # Parameters (kwargs)

    cwd:pathlib.Path
        - The path to change the current working directory too
        - Default - None

    # NOTE

    Reference: https://docs.python.org/3/library/subprocess.html

    >>> import shlex, subprocess
    >>> command_line = input()
    /bin/vikings -input eggs.txt -output "spam spam.txt" -cmd "echo '$MONEY'"
    >>> args = shlex.split(command_line)
    >>> print(args)
    ['/bin/vikings', '-input', 'eggs.txt', '-output', 'spam spam.txt', '-cmd', "echo '$MONEY'"]
    >>> p = subprocess.Popen(args) # Success!

    Using shlex and pasting the command into repl can make this process much easier to work with.

    """

    cwd = kwargs["cwd"] if "cwd" in kwargs else None

    p = subprocess.Popen(cmd, stdout=subprocess.PIPE, universal_newlines=True, cwd=cwd)

    # Gather the output from stdout to a list
    output = [l.strip() for l in p.stdout]

    # # Send the output to the logger
    # for l in output:
    #     log.info(l)

    return output



#!/usr/bin/env python3
# -*- coding:utf-8 -*-

# -----------
# SPDX-License-Identifier: MIT
# Copyright (c) 2021 Troy Williams

# uuid       = ff3c11de-b5b3-11eb-9fb7-a3fe2da49343
# author     = Troy Williams
# email      = troy.williams@bluebill.net
# date       = 2021-05-15
# -----------

"""
This module contains information for properly setting up a package that can
represent md_docs.
"""

# https://hynek.me/articles/testing-packaging/
# https://docs.python.org/3/distutils/setupscript.html


from setuptools import setup, find_packages
from pathlib import Path

README = Path(__file__).parent.joinpath("README.md").read_text()

setup(
    name="hindsight-ubuntu",
    version="0.0.2",
    description="Remember the position, size and virtual desktop of your open windows and restore them at a later point.",
    author="Troy Williams",
    author_email="troy.williams@bluebill.net",
    url="https://github.com/TroyWilliams3687/hindsight",
    long_description=README,
    long_description_content_type="text/markdown",
    license='MIT',
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.9",
        "Operating System :: POSIX :: Linux",
        "Topic :: Desktop Environment :: Gnome",
    ],
    python_requires=">=3.9",
    packages=find_packages(where="src"),
    package_dir={"": "src"},
    include_package_data=True,
    install_requires=[
        "click",
        "appdirs",
    ],
    entry_points={
        'console_scripts': [
            'hindsight=hindsight.hindsight:main',
        ],
    },
)



#!/usr/bin/env python3
# -*- coding:utf-8 -*-

# -----------
# SPDX-License-Identifier: MIT
# Copyright (c) 2021 Troy Williams

# uuid       = be569314-b7ee-11eb-89b8-59b62252b88f
# author     = Troy Williams
# email      = troy.williams@bluebill.net
# date       = 2021-05-18
# -----------

"""

Reference:
- https://askubuntu.com/questions/631392/saving-and-restoring-window-positions

"""

# ------------
# System Modules - Included with Python

import sys
import logging

from pathlib import Path

# ------------
# 3rd Party - From pip

import click
from appdirs import AppDirs

# ------------
# Custom Modules

from .save import save
from .restore import restore

# -------------
# Logging

# configure logging for this module - it is a bit different here because
# it is intended to be executed by the user <http://nathanielobrown.com/blog/posts/quick_and_dirty_python_logging_lesson.html>

# Logging Levels:
# CRITICAL
# ERROR
# WARNING
# INFO
# DEBUG
# NOTSET

# get the root logger
logger = logging.getLogger()
logger.setLevel(logging.INFO)  # change logging level here...

# make a console logger
console = logging.StreamHandler()

# add the console logger to the root logger
logger.addHandler(console)

# Assign the variable
log = logging.getLogger(__name__)
# -------------

__appname__ = "hindsight"
__company__ = "bluebill.net"


def common_paths():
    """
    The paths that the application will commonly use for storing settings
    and caching things.

    """

    dirs = AppDirs()

    paths = {
        "config": Path(dirs.user_config_dir)
        .joinpath(__company__)
        .joinpath(__appname__),
    }

    paths["config"].mkdir(parents=True, exist_ok=True)

    paths["locations"] = paths["config"].joinpath("locations.json")

    return paths


@click.group()
@click.version_option()
@click.pass_context
def main(*args, **kwargs):
    """

    # Parameters

    build_cfg:str
        - The path to the YAML configuration file to use to drive the process.

    # Usage


    """

    # Initialize the shared context object to a dictionary and configure it for the app
    ctx = args[0]
    ctx.ensure_object(dict)

    ctx.obj["paths"] = common_paths()


# -----------
# Add the child menu options

main.add_command(save)
main.add_command(restore)
